#include "cset.hh"

#include <sys/poll.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <vector>
#include <cerrno>
#include <cstdio>


/* Hardcoded parameters: */

static bool IsUTF8Channel(const std::string& channame)
{
    return channame == "#japani.fi";
}
static const std::string GetServer()
{
    return "irc.stealth.net";
}
static const std::string GetServerPort()
{
    return "6667";
}


/* Code */

const std::vector<std::string> Split(const std::string& text, char separator)
{
    std::vector<std::string> words;
    std::string rest = text;
    for(;;)
    {
        unsigned spacepos = rest.find(separator);  
        if(spacepos == rest.npos)break;
        words.push_back(rest.substr(0, spacepos));   
        if(spacepos+1 >= rest.size()) { rest = ""; break; }
        rest = rest.substr(spacepos+1);
    }
    if(rest.size()) { words.push_back(rest); rest = ""; }
    return words;
}


/* From client(utf8) to server */
static const std::string Latin1Out(const std::string& s)
{
    return toLatin1(fromUTF8(s));
}
static const std::string UTF8Out(const std::string& s)
{
    return toUTF8(fromUTF8(s));
}
static const std::string OutChannel(const std::string& chan, const std::string& line)
{
    return IsUTF8Channel(chan) ? UTF8Out(line) : Latin1Out(line);
}


/* From server to client(utf8) */
static const std::string Latin1In(const std::string& s)
{
    return toUTF8(fromLatin1(s));
}
static const std::string UTF8In(const std::string& s)
{
    return toUTF8(fromUTF8(s));
}
static const std::string InChannel(const std::string& chan, const std::string& line)
{
    return IsUTF8Channel(chan) ? UTF8In(line) : Latin1In(line);
}

static bool ReadWrite(FILE* inf, FILE* outf)
{
    char Buf[8192];
    if(!fgets(Buf, sizeof Buf, inf)) return true;
    
    std::string line = Buf;
    
    std::string guess_channel = "";
    
    enum { in, out } guess_dir = out;

    std::vector<std::string> words = Split(line, ' ');
    for(unsigned a=0; a<4; ++a)
    {
        if(a >= words.size()) break;
        const std::string& word = words[a];
        if(word.empty()) continue;
        
        char ch = word[0];
        if(a == 0 && ch == ':')
        {
            guess_dir = in; // from server
        }
        if(ch == '#' || ch == '&' || ch == '+')
        {
            guess_channel = word;
            break;
        }
    }
    
/*
    fprintf(stderr, "guess(chan=%s, dir=%s)\n",
        guess_channel.c_str(),
        guess_dir==in ? "in" : "out");
*/
    std::string result =
        guess_dir == in ?  InChannel(guess_channel, line)
                        : OutChannel(guess_channel, line);
    
    fwrite(result.c_str(), 1, result.size(), outf);
    return false;
}

static FILE* ConnectIRC()
{
    int pip[2];
    socketpair(AF_UNIX, SOCK_STREAM, 0, pip);
    if(!fork())
    {
        int p0 = dup(dup(pip[0]));
        dup2(p0, 0);
        dup2(p0, 1);
        for(int n=2; n<511; ++n) close(n);
        execlp("pcon",
               "pcon", GetServer().c_str(), GetServerPort().c_str(),
               "-i", "cat",
               "-o", "cat",
               NULL);
        _exit(-1);
    }
    close(pip[0]);
    return fdopen(pip[1], "r+");
}

int main(void)
{
    FILE* user  = fdopen(dup(0), "r+");
    fclose(stdin);
    fclose(stdout);
    FILE* server = ConnectIRC();
    
    // Buffering is not wanted
    setbuf(user,  NULL);
    setbuf(server, NULL);
    
    // Nonblocking is not wanted
    fcntl(fileno(user), F_SETFL,  fcntl(fileno(user), F_GETFL) & ~O_NONBLOCK);
    fcntl(fileno(server), F_SETFL, fcntl(fileno(server), F_GETFL) & ~O_NONBLOCK);
    
    for(;;)
    {
        struct pollfd polls[2] =
        {
            {fileno(user), POLLIN, 0},
            {fileno(server), POLLIN, 0}
        };
        int n = poll(polls, 2, -1);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(polls[0].revents & POLLIN)
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(polls[1].revents & POLLIN)
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
    
    fclose(user);
    fclose(server);
    
    return 0;
}
